from salesforce_tools.salesforce import SalesforceAPI, SalesforceModelFactory, RestAPI, ToolingAPI
from salesforce_tools.auth import login, SalesforceOAuthClient, SalesforceJWTClient
from salesforce_tools.bulk import BulkAPI, BulkJobException
from salesforce_tools.bulk_models import JobInfo, JobInfoList, JobTypeEnum, JobStateEnum, OperationEnum, ContentTypeEnum
from salesforce_tools.oauth_server import CallbackServer, OAuthCallbackHandler
from salesforce_tools.util import SFDateTime, EmailValidator, EMAIL_ADDRESS_REGEX



from salesforce_tools.async_sf.client import SalesforceAsyncOAuth2Client, SalesforceSfdxAsyncOAuth2Client, SalesforceAPISelector
from salesforce_tools.async_sf.metadata import SalesforceMetadataFetcherAsync



from authlib.integrations.httpx_client import AsyncOAuth2Client
import json
from salesforce_tools.auth import sfdx_auth_url_to_dict
from subprocess import check_output
from urllib.parse import quote


class SalesforceAsyncOAuth2Client(AsyncOAuth2Client):
    def __init__(self, *args, **kwargs):
        self.args = kwargs
        if kwargs.get('instance_url'):
            self.instance_url = kwargs.get('instance_url')
        elif kwargs.get('token') and kwargs.get('instance_url'):
            self.instance_url = kwargs.get('token', {}).get('instance_url')
        else:
            default_auth_url = 'https://test.salesforce.com' if kwargs.get(
                'sandbox') else 'https://login.salesforce.com'
            self.instance_url = kwargs.get('auth_url', default_auth_url)
        self.api_version = str(kwargs.get('api_version', '56.0'))
        client_args = {k: v for k, v in kwargs.items() if k not in ['instance_url', 'api_version', 'sandbox']}
        if self.instance_url and self.api_version and not kwargs.get('base_url'):
            client_args['base_url'] = f"{self.instance_url}/services/data/v{self.api_version}/"
        client_args['token_endpoint'] = f"{self.instance_url}/services/oauth2/token"
        super().__init__(**client_args)
        self.register_compliance_hook('access_token_response', self._fix_token_response)
        self.register_compliance_hook('refresh_token_response', self._fix_token_response)

    def query(self, qry):
        return self.get(f'query?q={quote(qry)}')

    def _fix_token_response(self, resp):
        data = resp.json()
        data['expires_in'] = 3600
        resp.json = lambda: data
        return resp


class SalesforceSfdxAsyncOAuth2Client(SalesforceAsyncOAuth2Client):
    def __init__(self, alias, **kwargs):
        sfdx_auth_url = check_output(f'sfdx force:org:display -u {alias} --json --verbose', shell=True)
        sfdx_auth_url = json.loads(sfdx_auth_url.decode('utf-8'))
        kwargs.update(sfdx_auth_url_to_dict(sfdx_auth_url['result']['sfdxAuthUrl']))
        kwargs['token']['access_token'] = sfdx_auth_url['result']['accessToken']
        super().__init__(**kwargs)


class SalesforceAPISelector():
    def __init__(self, *args, **kwargs):
        self.sf = SalesforceAsyncOAuth2Client(**kwargs)
        self._oauth_user_info_url = f"{self.sf.instance_url}/services/oauth2/userinfo"
        self._userinfo = None
        self.api_version = self.sf.api_version

    @property
    async def userinfo(self):
        if not self._userinfo:
            self._userinfo = (await self.sf.get(self._oauth_user_info_url)).json()
        return self._userinfo

    @property
    async def apis(self):
        return (await self.userinfo).get('urls')

    async def __getattr__(self, item):
        base_url = (await self.userinfo).get('urls').get(item+'_rest').replace('{version}', self.api_version)
        args = self.sf.args
        args['base_url'] = base_url
        return SalesforceAsyncOAuth2Client(**args)



import asyncio
from salesforce_tools.async_sf.client import SalesforceAsyncOAuth2Client


class SalesforceMetadataFetcherAsync:
    def __init__(self, client: SalesforceAsyncOAuth2Client, cache_file: str = None):
        self.client = client
        self._cache_sobject = {}
        self._cache_sobjects_list = []
        self.cache_file = cache_file
        if self.cache_file:
            self.load_cache(self.cache_file)

    async def get_sobject_describe(self, obj, cache=True):
        if not cache or not self._cache_sobject.get(obj):
            self._cache_sobject[obj] = (await self.client.get(f'sobjects/{obj}/describe', timeout=10.0)).json()
        return self._cache_sobject.get(obj)

    async def get_picklist_values(self, obj, field):
        obj_md = await self.get_sobject_describe(obj)
        return [f for f in obj_md['fields'] if f['name'] == field][0]['picklistValues']

    async def get_permissionable_fields(self, normalize=True):
        permissionable_fields = {}
        pv = await self.get_picklist_values('FieldPermissions', 'Field')
        pv = [p['value'] for p in pv]
        if normalize:
            for f in pv:
                sf_obj, fld = f.split('.')
                if not permissionable_fields.get(sf_obj):
                    permissionable_fields[sf_obj] = []
                permissionable_fields[sf_obj].append(fld)
            return permissionable_fields
        return pv

    async def get_all_sobjects(self, cache=True):
        if not cache or not self._cache_sobjects_list:
            self._cache_sobjects_list = (await self.client.get("sobjects", timeout=10.0)).json()['sobjects']
        return self._cache_sobjects_list

    async def get_all_sobject_request_coroutines(self, unfiltered=True, cache=True):
        objects_to_fetch = await self.get_all_sobjects(cache)
        tasks = []
        if not unfiltered:
            objects_to_fetch = [o for o in objects_to_fetch if
                                o['associateEntityType'] not in ['Share', 'ChangeEvent', 'Feed', 'History']]
        if cache:
            objects_to_fetch = [o for o in objects_to_fetch if o['name'] not in self._cache_sobject.keys()]
        if objects_to_fetch:
            tasks = [self.get_sobject_describe(o['name']) for o in objects_to_fetch]
        return tasks

    async def get_all_sobject_metadata(self, unfiltered=True, cache=True):
        tasks = await self.get_all_sobject_request_coroutines(unfiltered, cache)
        [await t for t in asyncio.as_completed(tasks)]
        return self._cache_sobject

    def save_cache(self, filename: str = None):
        filename = filename or self.cache_file
        output_format = {"_cache_sobjects_list": self._cache_sobjects_list,
                         "_cache_sobject": self._cache_sobject}
        with open(filename, 'w') as f:
            json.dump(output_format, f, indent=4)

    def load_cache(self, filename: str = None):
        filename = filename or self.cache_file
        try:
            with open(filename, 'r') as f:
                c = json.load(f)
                self._cache_sobjects_list, self._cache_sobject = c['_cache_sobjects_list'], c['_cache_sobject']
        except FileNotFoundError:
            pass



from pydantic import BaseModel, ValidationError, validator, Field
from typing import Literal, Optional, List
from enum import Enum
import datetime


class JobStateEnum(str, Enum):
    Open = 'Open'
    Closed = 'Closed'
    Aborted = 'Aborted'
    Failed = 'Failed'
    UploadComplete = 'UploadComplete'


class ContentTypeEnum(str, Enum):
    CSV = 'CSV'
    JSON = 'JSON'
    XML = 'XML'
    ZIP_CSV = 'ZIP_CSV'
    ZIP_JSON = 'ZIP_JSON'
    ZIP_XML = 'ZIP_XML'


class ContentTypeHeaderEnum(str, Enum):
    CSV = 'text/csv'
    JSON = 'application/json'
    XML = 'application/xml'


class ConcurrencyModeEnum(str, Enum):
    Parallel = 'Parallel'
    Serial = 'Serial'


class OperationEnum(str, Enum):
    upsert = 'upsert'
    update = 'update'
    insert = 'insert'
    delete = 'delete'
    hard_delete = 'hardDelete'
    query = 'query'
    query_all = 'queryAll'


class ColumnDelimiterEnum(str, Enum):
    BACKQUOTE = "BACKQUOTE"
    CARET = "CARET"
    COMMA = "COMMA"
    PIPE = "PIPE"
    SEMICOLON = "SEMICOLON"
    TAB = "TAB"


class LineEndingEnum(str, Enum):
    CRLF = 'CRLF'
    LF = 'LF'


class JobTypeEnum(str, Enum):
    V2Ingest = "V2Ingest"
    Classic = "Classic"
    BigObjectIngest = "BigObjectIngest"


class BatchStateEnum(str, Enum):
    Queued = "Queued"
    InProgress = "InProgress"
    Completed = "Completed"
    Failed = "Failed"
    NotProcessed = "NotProcessed"


class JobInfo(BaseModel):
    api_version: Optional[float] = Field(alias='apiVersion')
    apex_processing_time: Optional[int] = Field(alias='apexProcessingTime')
    api_active_processing_time: Optional[int] = Field(alias='apiActiveProcessingTime')
    assignment_rule_id: Optional[str] = Field(alias='assignmentRuleId')
    concurrency_mode: Optional[ConcurrencyModeEnum] = Field(alias='concurrencyMode')
    content_type: ContentTypeEnum = Field(alias='contentType', default=ContentTypeEnum.CSV)
    created_by_id: Optional[str] = Field(alias='createdById')
    created_date: Optional[datetime.datetime] = Field(alias='createdDate')
    external_id_field_name: Optional[str] = Field(alias='externalIdFieldName')
    id: Optional[str]
    number_batches_completed: Optional[int] = Field(alias='numberBatchesCompleted')
    number_batches_queued: Optional[int] = Field(alias='numberBatchesQueued')
    number_batches_failed: Optional[int] = Field(alias='numberBatchesFailed')
    number_batches_in_progress: Optional[int] = Field(alias='numberBatchesInProgress')
    number_batches_total: Optional[int] = Field(alias='numberBatchesTotal')
    number_records_failed: Optional[int] = Field(alias='numberRecordsFailed')
    number_records_processed: Optional[int] = Field(alias='numberRecordsProcessed')
    number_retries: Optional[int] = Field(alias='numberRetries')
    sobject: Optional[str] = Field(alias='object')
    operation: OperationEnum
    state: Optional[JobStateEnum]
    column_delimiter: Optional[ColumnDelimiterEnum] \
        = Field(alias='columnDelimiter')
    line_ending: Optional[LineEndingEnum] = Field(alias='lineEnding')
    job_type: Optional[JobTypeEnum] = Field(alias='jobType', default=JobTypeEnum.Classic)
    total_processing_time: Optional[int] = Field(alias='totalProcessingTime')
    systemModstamp: Optional[datetime.datetime] = Field(alias='systemModstamp')
    content_url: Optional[str] = Field(alias='contentUrl')
    query: Optional[str]


    @validator('operation')
    def external_id_field_name_required_for_upsert(cls, v, values, **kwargs):
        if v == 'upsert' and not values.get('external_id_field_name'):
            raise ValueError('External Id Field Name is required for upsert')
        if not values.get('sobject') and v != 'query':
            raise ValueError('Object must be specified')
        return v

    class Config:
        allow_population_by_field_name = True


class BatchInfo(BaseModel):
    apex_processing_time: Optional[int] = Field(alias='apexProcessingTime')
    api_active_processing_time: Optional[int] = Field(alias='apiActiveProcessingTime')
    created_date: Optional[datetime.datetime] = Field(alias='createdDate')
    id: Optional[str]
    job_id: Optional[str] = Field(alias='jobId')
    number_records_failed: Optional[int] = Field(alias='numberRecordsFailed')
    number_records_processed: Optional[int] = Field(alias='numberRecordsProcessed')
    state: Optional[BatchStateEnum]
    state_message: Optional[str] = Field(alias='stateMessage')
    system_modstamp: Optional[datetime.datetime] = Field(alias='systemModstamp')
    total_processing_time: Optional[int] = Field(alias='totalProcessingTime')


class BatchInfoList(BaseModel):
    records: List[BatchInfo] = Field(alias='batchInfo', default=[])

    class Config:
        allow_population_by_field_name = True


class JobInfoList(BaseModel):
    records: List[JobInfo] = Field(alias='jobInfo', default=[])

    class Config:
        allow_population_by_field_name = True


class APIError(BaseModel):
    code: str = Field(alias='errorCode')
    message: str = Field(alias='message')

    class Config:
        allow_population_by_field_name = True


class BulkAPIError(BaseModel):
    code: str = Field(alias='exceptionCode')
    message: str = Field(alias='exceptionMessage')

    class Config:
        allow_population_by_field_name = True

class BulkException(Exception):
    error: BulkAPIError


class ExceptionCode(str, Enum):
    ClientInputError = 'ClientInputError'
    ExceededQuota = 'ExceededQuota'
    FeatureNotEnabled = 'FeatureNotEnabled'
    InvalidBatch = 'InvalidBatch'
    InvalidJob = 'InvalidJob'
    InvalidJobState = 'InvalidJobState'
    InvalidOperation = 'InvalidOperation'
    InvalidSessionId = 'InvalidSessionId'
    InvalidUrl = 'InvalidUrl'
    InvalidUser = 'InvalidUser'
    InvalidXML = 'InvalidXML'
    Timeout = 'Timeout'
    TooManyLockFailure = 'TooManyLockFailure'
    Unknown = 'Unknown'



import json
from requests_oauthlib import OAuth2Session
from oauthlib.common import to_unicode
from oauthlib.oauth2 import WebApplicationClient, BackendApplicationClient
from oauthlib.oauth2.rfc6749.parameters import prepare_token_request
from datetime import datetime, timedelta
import webbrowser
from salesforce_tools.oauth_server import CallbackServer
from urllib.parse import urlsplit, urljoin
import os
from typing import Callable
import jwt
from urllib.parse import urlparse


os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = "1"
AUTH_URL = 'https://login.salesforce.com'
AUTH_REL_URL = '/services/oauth2/authorize'
TOKEN_REL_URL = '/services/oauth2/token'
TOKEN_LIFE = timedelta(hours=2)


def login(client_id: str = None, client_secret: str = None, token: dict = {},
          token_updater: Callable = lambda x: True, callback_port: int = 8000,
          force_login: bool = False, scope: str = 'refresh_token openid web full',
          auth_url: str = AUTH_URL, redirect_url: str = None, private_key: str = None,
          private_key_filename: str = None, username: str = None, **kwargs):
    auth_code_url = urljoin(auth_url, AUTH_REL_URL)
    token_url = urljoin(auth_url, TOKEN_REL_URL)
    redirect_url = redirect_url or f"http://localhost:{callback_port}/callback"
    if private_key or private_key_filename:
        client = SalesforceJWTClient(client_id, username=username, auth_url=auth_url,
                                     private_key_filename=private_key_filename, private_key=private_key)
        salesforce = salesforce_compliance_fix(
            SalesforceOAuth2Session(client=client, auto_refresh_url=token_url, token_updater=token_updater)
        )
        salesforce.fetch_token(token_url)
        if token_updater:
            token_updater(salesforce.token)
    else:
        salesforce = salesforce_compliance_fix(
            SalesforceOAuth2Session(token=token,
                          client=SalesforceOAuthClient(client_id),
                          redirect_uri=redirect_url,
                          scope=scope if not token else None,
                          auto_refresh_url=token_url,
                          auto_refresh_kwargs={k: v for k, v in
                                               {'client_id': client_id,
                                                'client_secret': client_secret}.items()
                                               if v},
                          token_updater=token_updater
                          )
        )
        if force_login or not token or not token.get('refresh_token'):
            authorization_url, state = salesforce.authorization_url(auth_code_url)
            webbrowser.open(authorization_url, new=1)
            authorization_response = CallbackServer().get_auth(port=callback_port)
            ruri = urlsplit(redirect_url)
            ruri_base_url = ruri.scheme + '://' + ruri.netloc
            authorization_response = urljoin(ruri_base_url, authorization_response)
            salesforce.fetch_token(token_url, client_secret=client_secret,
                                   authorization_response=authorization_response)
            if token_updater:
                token_updater(salesforce.token)
    return salesforce


def salesforce_compliance_fix(sess):
    token = ''

    def _compliance_fix(response):
        token = response.json()
        if token.get('issued_at'):
            iat = int(token["issued_at"]) / 1000
            token["expires_in"] = (datetime.fromtimestamp(iat) + TOKEN_LIFE - datetime.now()).seconds
        else:
            token["expires_in"] = TOKEN_LIFE.seconds
        fixed_token = json.dumps(token)
        response._content = to_unicode(fixed_token).encode("utf-8")
        if token.get('instance_url'):
            sess.instance_url = token.get('instance_url')
        return response

    sess.register_compliance_hook("access_token_response", _compliance_fix)
    sess.register_compliance_hook("refresh_token_response", _compliance_fix)

    return sess


class SalesforceOAuthClient(WebApplicationClient):
    def _add_bearer_token(self, uri, http_method='GET', body=None,
                          headers=None, token_placement=None):
        uri, headers, body = super()._add_bearer_token(
            uri,
            http_method=http_method,
            body=body,
            headers=headers,
            token_placement=token_placement
        )

        headers['X-SFDC-Session'] = self.token.get('access_token')
        return uri, headers, body


class SalesforceJWTClient(BackendApplicationClient):
    grant_type = 'urn:ietf:params:oauth:grant-type:jwt-bearer'

    def __init__(self, client_id: str, username: str,
                 auth_url: str = 'https://test.salesforce.com',
                 private_key_filename: str = None,
                 private_key: str = None,
                 **kwargs):
        super().__init__(client_id, **kwargs)
        self.audience = auth_url
        self.private_key = private_key
        self.user_name = username
        if private_key_filename:
            with open(private_key_filename, mode='r') as f:
                self.private_key = ''.join(f.readlines())

    def prepare_refresh_body(self, body: str = '', **kwargs):
        return self.prepare_request_body(body)

    def prepare_request_body(self, body: str = '', **kwargs):
        claims = {"iss": self.client_id,
                  "sub": self.user_name,
                  "aud": self.audience,
                  "exp": int(datetime.now().timestamp()) + 60 * 3
                  }

        signed_claims = jwt.encode(claims, self.private_key, "RS256")
        return prepare_token_request(self.grant_type, assertion=signed_claims, format="json", **kwargs)

    def _add_bearer_token(self, uri, http_method='GET', body=None,
                          headers=None, token_placement=None):
        uri, headers, body = super()._add_bearer_token(
            uri,
            http_method=http_method,
            body=body,
            headers=headers,
            token_placement=token_placement
        )

        headers['X-SFDC-Session'] = self.token.get('access_token')
        return uri, headers, body


def sfdx_auth_url_to_dict(url):
    p_url = urlparse(url)
    if p_url.scheme == 'force':
        i_url = f"https://{p_url.hostname}"
        return {
            "client_id": p_url.username,
            "client_secret": p_url.password.split(':')[0],
            "token": {
                "refresh_token": p_url.password.split(':')[1],
                "instance_url": i_url,
                "access_token": 'REFRESH_ME',
                "expires_at": int(datetime.now().timestamp())

            },
            "auth_url": i_url

        }


class SalesforceOAuth2Session(OAuth2Session):
    def __init__(self, instance_url=None, api_root=None, *args, **kwargs):
        super(SalesforceOAuth2Session, self).__init__(*args, **kwargs)
        self.instance_url = instance_url
        if not self.instance_url and self.token and self.token.get('instance_url'):
            self.instance_url = self.token.get('instance_url')
        self.api_root = api_root

    def request(self, method, url, *args, **kwargs):
        if not self.instance_url and self.token and self.token.get('instance_url'):
            self.instance_url = self.token.get('instance_url')
        base_url = ''.join(filter(None, [self.instance_url, self.api_root]))
        url = urljoin(base_url, url)
        return super(SalesforceOAuth2Session, self).request(method, url, *args, **kwargs)




import urllib.parse
from salesforce_tools.salesforce import SalesforceAPI
from enum import Enum
from salesforce_tools.bulk_models import JobInfo, JobInfoList, BatchInfo, BatchInfoList, \
    OperationEnum, ContentTypeEnum, ContentTypeHeaderEnum, JobTypeEnum, JobStateEnum, BulkAPIError, APIError,\
    BulkException
from typing import Union, Optional, List
from pydantic import BaseModel, ValidationError, parse_obj_as


class BulkJobException(Exception):
    pass


def get_enum_or_val(e):
    return e.value if isinstance(e, Enum) else e


class BulkAPI(SalesforceAPI):
    job: Optional[Union[JobInfo, BulkAPIError]]
    batches: List[BatchInfo] = []

    @property
    def job(self):
        return self.__job

    @job.setter
    def job(self, value: JobInfo):
        if value and isinstance(value, JobInfo):
            value.job_type = value.job_type or (JobTypeEnum.V2Ingest
                                                if value.content_url and 'ingest' in value.content_url
                                                else JobTypeEnum.Classic)
            self.job_type = value.job_type
        else:
            self.job_type = JobTypeEnum.Classic
        self.__job = value

    def __init__(self, job_id: str = None, job: JobInfo = None, **kwargs):
        super().__init__(**kwargs)
        self.job = kwargs.get('job', self.set_job(job_id) if job_id else job)

    def _get_results_url(self, job_type=None, operation=None):
        if isinstance(self.job, JobInfo):
            job_type = JobTypeEnum.V2Ingest if job_type == JobTypeEnum.V2Ingest else JobTypeEnum.Classic
            operation = operation or self.job.operation
            if operation == "query":
                return f"/services/data/v{self.api_version}/jobs/query/{self.job.id}/results"
            if job_type == JobTypeEnum.V2Ingest:
                return f"/services/data/v{self.api_version}/jobs/ingest/{self.job.id}/results"
            else:
                return f'/services/async/{self.api_version}/job/{self.job.id}/batch'

    def _get_job_url(self, job_type=None, operation=None):
        job_type = job_type or self.job_type
        operation = operation or self.job.operation if isinstance(self.job, JobInfo) else None
        url = f'/services/data/v{self.api_version}/jobs/ingest' if job_type == JobTypeEnum.V2Ingest else \
            f'/services/async/{self.api_version}/job'
        if operation == 'query':
            url = f'/services/data/v{self.api_version}/jobs/query'
        else:
            return f'/services/async/{self.api_version}/job'
        return url

    def create_job(self, job: JobInfo):
        self.job = job
        job_type = JobTypeEnum.V2Ingest if job.job_type == JobTypeEnum.V2Ingest else JobTypeEnum.Classic
        url = self._get_job_url()

        if not job.id:
            d = job.dict(by_alias=True, exclude_none=True, exclude={'job_type'})
            job, ok, *_ = self.request(url, method='POST', json=d)
            self.job = self._model_wrap(job, ok, JobInfo)
            if ok:
                self.job.job_type = job_type
        return self.job

    # TODO: Where to put "static helper methods"?
    def get_jobs(self):
        jobs, ok, *_ = self.request(f'/services/data/v{self.api_version}/jobs/ingest')
        return self._model_wrap(jobs, ok, JobInfoList, False)

    def upload_data(self, data):
        if isinstance(self.job, BulkAPIError):
            raise BulkException(self.job)
        job_id = self.job.id
        content_type = getattr(ContentTypeHeaderEnum, self.job.content_type).value
        url = f'/services/async/{self.api_version}/job/{job_id}/batch'
        data, ok, *_ = self.request(url, method='POST', data=data, headers={'Content-Type': content_type})
        batch = self._model_wrap(data, ok, BatchInfo, False)
        self.batches.append(batch)
        return batch

    def _model_wrap(self, data: any, ok: bool, model: BaseModel, raise_exception_on_error=False):
        if ok:
            o = parse_obj_as(model, data)
        else:
            if isinstance(data, list):
                try:
                    o = parse_obj_as(List[BulkAPIError], data)
                except ValidationError:
                    o = parse_obj_as(List[APIError], data)
            else:
                if data.get('error'):
                    data = data.get('error')
                try:
                    o = parse_obj_as(BulkAPIError, data)
                except Exception:
                    o = parse_obj_as(APIError, data)

            if raise_exception_on_error:
                raise ValueError(o)
        return o

    def close_job(self, job_id: str = None, state: JobStateEnum = JobStateEnum.UploadComplete):
        job_id = job_id or self.job.id
        url = f'/services/async/{self.api_version}/job/{job_id}'
        data, ok, *_ = self.request(url, method='POST', json={"state": state.value})
        job = self._model_wrap(data, ok, JobInfo, False)
        if ok and job.id == self.job.id:
            self.job = job
        return job

    def set_job(self, job_id, job_type=JobTypeEnum.Classic, operation=None):
        url = f"{self._get_job_url(job_type, operation)}/{job_id}"
        data, ok, *_ = self.request(url, method='GET')
        self.job = self._model_wrap(data, ok, JobInfo, False)
        return self.job

    def get_batch_info(self, batch: BatchInfo = None, batch_id: str = None):
        url = f'/services/async/{self.api_version}/job/{self.job.id}/batch'
        data, ok, *_ = self.request(url, method='GET')
        return self._model_wrap(data, ok, BatchInfoList, False)

    def get_batches(self, job_id: str = None):
        url = f'/services/async/{self.api_version}/job/{self.job.id}/batch'
        data, ok, *_ = self.request(url, method='GET')
        self.batches = self._model_wrap(data, ok, BatchInfoList, True).records
        return self.batches

    # Sforce-Locator, Sforce-NumberOfRecords, Sforce-Limit-Info
    def get_results(self, locator=None, max_records=None):
        params = {'locator': locator, 'maxRecords': max_records}
        params = {k: v for k, v in params.items() if v}
        url = urllib.parse.urljoin(self._get_results_url(), '?' + urllib.parse.urlencode(params))
        return self.request(url)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['salesforce_tools', 'salesforce_tools.async_sf']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.2.1,<3.0.0',
 'PyJWT>=2.4.0,<3.0.0',
 'authlib>=1.1.0,<2.0.0',
 'cryptography>=38.0.3,<39.0.0',
 'flatten-json>=0.1.13,<0.2.0',
 'httpx>=0.23.3,<0.24.0',
 'pandas>=1.4.3,<2.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'requests-oauthlib>=1.3.1,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'xmltodict>=0.13.0,<0.14.0']

setup_kwargs = {
    'name': 'salesforce-tools',
    'version': '0.1.30',
    'description': 'Salesforce API tools',
    'long_description': 'None',
    'author': 'David Manuel',
    'author_email': 'david@dcmanjr.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/dacmanj/salesforce-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
